<?php
/**
 * Бэкенд «BIFIT»
 *
 * @package    DIAFAN.CMS
 * @author     https://user.diafan.ru/user/eamat/
 * @version    6.0
 * @license    http://www.diafan.ru/license.html
 * @copyright  Copyright (c) 2003-2023 OOO «Диафан» (http://www.diafan.ru/)
 */

if (!defined("DIAFAN")) {
	$path = __FILE__;

	while (!file_exists($path . "/includes/404.php")) {
		$parent = dirname($path);

		if ($parent == $path) {
			exit();
		}

		$path = $parent;
	}

	include $path . "/includes/404.php";
}

class Cashregister_bifit extends Diafan
{
	/**
	 * Получение переменных из конфигурации модуля
	 *
	 * @param string $name
	 *
	 * @return string|bool
	 */
	public function __get($name)
	{
		return $this->diafan->configmodules("bifit_" . $name, "cashregister");
	}

	/**
	 * Создание переменных в конфигурации модуля
	 *
	 * @param string $name
	 * @param mixed  $value
	 *
	 * @return string
	 */
	public function __set($name, $value)
	{
		return $this->diafan->configmodules(
			"bifit_" . $name,
			"cashregister",
			false,
			false,
			$value
		);
	}

	/**
	 * Чек «Полная оплата»
	 *
	 * @param array $info данные о заказе
	 *
	 * @return sting  Уникальный идентификатор чека
	 */
	public function sell($info)
	{
		return $this->createReceipt($info, "sell", "full_payment");
	}

	/**
	 * Чек «Предоплата 100%»
	 *
	 * @param array $info данные о заказе
	 *
	 * @return sting  Уникальный идентификатор чека
	 */
	public function presell($info)
	{
		return $this->createReceipt($info, "sell", "full_prepayment");
	}

	/**
	 * Чек «Возврат прихода»
	 *
	 * @param array $info данные о заказе
	 *
	 * @return sting Уникальный идентификатор чека
	 */
	public function refund($info)
	{
		return $this->createReceipt($info, "sell_refund");
	}

	/**
	 * POST запрос для чеков
	 *
	 * @param array  $info           данные о заказе
	 * @param enum string $operation
	 * @param string $payment_method способ расчёта
	 *
	 * @return sting Уникальный идентификатор чека
	 * @throws BifitException
	 */
	private function createReceipt($info, $operation, $payment_method = "")
	{
		// параметры
		$log_ok = $this->diafan->configmodules("bifit_log", "cashregister");
		$test = $this->diafan->configmodules("bifit_test", "cashregister")
			? "-test"
			: "";
		$token =
			$test != ""
				? "P5cKbUUD9uSSrSlGdzspLblvBnD0GzTAE0cLmAPSEMxJ79DtLE"
				: $this->diafan->configmodules("bifit_token", "cashregister");

		if ($token != "") {
			// запускаем БИФИТ-процесс
			self::__mainProcess($info, $test, $token, $log_ok, "front");
		}

		return 0;
	}

	public static function __createGUID($order)
	{
		// Create GUID (Globally Unique Identifier)
		$guid = "";
		$namespace = rand(11111, 99999);
		$uid = uniqid("", true);
		$data = $namespace;
		$data .= $_SERVER["REQUEST_TIME"];
		$data .= $_SERVER["HTTP_USER_AGENT"];
		$data .= $_SERVER["REMOTE_ADDR"];
		$data .= $_SERVER["REMOTE_PORT"];
		$data .= $order["number"];
		$data .= $order["summ"];
		$hash = strtoupper(hash("ripemd128", $uid . $guid . md5($data)));
		$guid =
			substr($hash, 0, 8) .
			"-" .
			substr($hash, 8, 4) .
			"-" .
			substr($hash, 12, 4) .
			"-" .
			substr($hash, 16, 4) .
			"-" .
			substr($hash, 20, 12);

		return strtolower($order["cms"] . "-" . $guid);
	}

	function __checkResponce($response)
	{
		$mess = "";

		if (
			is_object($response) &&
			($response->type == "ERROR" || $response->status == 400)
		) {
			$mess = "";

			if ($response->status == 400) {
				$mess .= $response->error . "\n";
				$mess .= $response->message . "\n";
			} else {
				$array = json_decode(json_encode($response), true);
				array_walk_recursive($array, function ($item, $key) use (
					&$mess
				) {
					if ($item != "ERROR" && trim($item != "")) {
						$mess .= $key . ": " . $item . "\n";
					}
				});
			}
		} elseif (is_object($response)) {
			$array = json_decode(json_encode($response), true);
			array_walk_recursive($array, function ($item, $key) use (&$mess) {

				if (trim($item != "")) {
					$mess .= $key . ": " . $item . "\n";
				}
			});
		}

		return $mess;
	}

	function __mainProcess($info, $test, $token, $log_ok, $loc)
	{
		// 1. АВТОРИЗАЦИЯ
		//    процесс необходимый для подключения клиента к фискальному процессингу

		// адрес для POST-запроса
		$url = "https://fp" . $test . ".bifit.com/processing-api/oauth/token";

		// данные ссылки
		$data = "token=" . $token;
		$data .= "&client_id=processing-connector-token";
		$data .= "&client_secret=processing-connector-token";
		$data .= "&grant_type=token";

		// headers
		$headers = ["Content-Type: application/x-www-form-urlencoded"];

		// декодируем полученный ответ
		$bifit = json_decode(self::__loadData($url, $headers, $data));

		// 2. ФОРМИРОВАНИЕ КОНТЕНТА ЧЕКА
		//    формирование информации для дальнейшей отправки на фискализацию

		// параметры плагина
		$cashier_name = $this->diafan->configmodules(
			"bifit_cashier_name",
			"cashregister"
		);
		$cashier_inn = $this->diafan->configmodules(
			"bifit_cashier_inn",
			"cashregister"
		);
		$receipt_type = $this->diafan->configmodules(
			"bifit_receipt_type",
			"cashregister"
		);
		$tax_system = $this->diafan->configmodules(
			"bifit_tax_system",
			"cashregister"
		);
		$calc_method = $this->diafan->configmodules(
			"bifit_calculation_method",
			"cashregister"
		);
		//$vat_mode = $this->diafan->configmodules('bifit_vat_mode', 'cashregister');
		$vat = $this->diafan->configmodules("bifit_vat", "cashregister");
		$payment_type = $this->diafan->configmodules(
			"bifit_payment_type",
			"cashregister"
		);
		$payment_address = $this->diafan->configmodules(
			"bifit_payment_address",
			"cashregister"
		);
		$address_type = $this->diafan->configmodules(
			"bifit_payment_address",
			"cashregister"
		);
		//$fio = $this->diafan->configmodules('bifit_fio', 'cashregister');

		// адрес клиента в чеке
		if ($address_type == "phone" && $info["phone"]) {
			$address = self::__checkPhone($info["phone"]);
		} else {
			$address = self::__checkEmail($info["email"]);
		}

		// параметры массива
		$o_subtotal = 0;
		$r_products = [];
		$products = $info["rows"];
		$o_shipment = $info["delivery_summ"];

		// массив товаров заказа
		foreach ($products as $key_id => $product) {
			//// корректируем цену товара с учётом скидок
			$p_price = self::__formatprc(
				self::__checkProductPrice(
					$product["price"],
					$info["summ_goods"] + $info["tax"],
					$info["discount_summ"],
					0
				)
			);
			$o_subtotal += $p_price * $product["count"];

			// очищаем имя товара от спец-символов
			$p_name = preg_replace(
				"/[^\p{L}0-9 ]/iu",
				" ",
				html_entity_decode($product["name"])
			);
			$p_name = preg_replace("/ +/", " ", trim($p_name));

			// формируем массив товаров для чека
			$r_products[$key_id] = new stdClass();
			$r_products[$key_id]->product_name = $p_name;
			$r_products[$key_id]->product_item_price = $p_price;
			$r_products[$key_id]->product_quantity = $product["count"];
			$r_products[$key_id]->product_tax = 0;
		}

		// сверка итоговых сумм заказа и чека
		$total_1 = self::__formatprc($o_subtotal + $o_shipment);
		$total_2 = self::__formatprc($info["summ"]);
		$diff = $total_1 != $total_2 ? abs($total_1 - $total_2) : 0;

		// параметры для чека
		$items = "";

		// формируем данные чека
		foreach ($r_products as $key_id => $product) {
			// цена товара
			$p_price = $product->product_item_price;

			// изменение цены товара согласно сверки
			if ($diff > 0 && $p_price > $diff) {
				$p_price =
					$total_2 > $total_1
						? $p_price + $diff / $product->product_quantity
						: $p_price - $diff / $product->product_quantity;
				$diff = 0;
			}

			// выводим данные товара в чек
			$items .= $items != "" ? ", " : "";
			$items .= "{";
			$items .= '"calculationMethod":"' . $calc_method . '", ';
			$items .= '"paymentSubject":"PRODUCT", ';
			$items .= '"name":"' . $product->product_name . '", ';
			$items .= '"price":' . self::__formatprc($p_price) . ", ";
			$items .= '"quantity":' . $product->product_quantity . ", ";
			//$items .= $vat_mode == 1 ? '"vat":"' . $vat . '", ' : '"vat":"VAT_' . floor($product->product_tax) . '", ';
			$items .= '"vat":"' . $vat . '", ';
			$items .=
				'"total":' .
				self::__formatprc($p_price * $product->product_quantity);
			$items .= "}";
		}

		// доставка
		if ($o_shipment > 0) {
			// выводим данные способа доставки в чек
			$items .= $items != "" ? ", " : "";
			$items .= "{";
			$items .= '"calculationMethod":"' . $calc_method . '", ';
			$items .= '"paymentSubject":"SERVICE", ';
			$items .= '"name":"Доставка", ';
			$items .= '"price":' . self::__formatprc($o_shipment) . ", ";
			$items .= '"quantity":1, ';
			$items .= '"vat":"' . $vat . '", ';
			$items .= '"total":' . self::__formatprc($o_shipment);
			$items .= "}";
		}

		// формируем чек
		$receipt = "{";
		$receipt .=
			'"type":"' .
			$receipt_type .
			'", "taxSystem":"' .
			$tax_system .
			'", ';
		$receipt .= '"cashier":{';
		$receipt .= '"name":"' . $cashier_name . '"';
		$receipt .=
			$cashier_inn != ""
				? ', "inn":"' . self::__checkInn($cashier_inn) . '"'
				: "";
		$receipt .= "}, ";
		$receipt .= '"client":{';
		$receipt .= '"address":"' . $address . '"';
		$receipt .= "}, ";
		$receipt .= '"paymentAddress":"' . $payment_address . '", ';
		$receipt .= '"items":[' . $items . "], ";
		$receipt .= '"total":' . self::__formatprc($info["summ"]) . ", ";
		$receipt .=
			'"payments":{"' .
			$payment_type .
			'": ' .
			self::__formatprc($info["summ"]) .
			"}";
		$receipt .= "}";

		// 3. ОТПРАВКА ЧЕКА В ФИСКАЛЬНЫЙ ПРОЦЕССИНГ
		//    передача контента чека в фискальный процессинг для дальнейшей фискализации

		// адрес для POST-запроса
		$url =
			"https://fp" .
			$test .
			".bifit.com/processing-api/protected/documents/registration/receipts";

		// данные чека
		$data = $receipt;

		// ключ идемпотентности (Idempotency-Key)
		$o = [];
		$o["number"] = $info["cashregister_id"];
		$o["summ"] = self::__formatprc($info["summ"]);
		$o["cms"] = "Diafan";
		$idemp_key = self::__createGUID($o);

		// headers
		$headers = [
			"Content-Type: application/json",
			"Authorization: Bearer " . $bifit->access_token,
			"Idempotency-Key: " . $idemp_key,
		];

		// получаем идентификатор документа на процессинге (id)
		$receipt_id = json_decode(self::__loadData($url, $headers, $data));

		// 4. ПОЛУЧЕНИЕ ДОКУМЕНТА ПО ID
		//    получение документа с фискальными признаками по его номеру

		// параметры
		$doc = "";

		if (!is_object($receipt_id)) {
			// ссылка для GET-запроса
			$url =
				"https://fp" .
				$test .
				".bifit.com/processing-api/protected/documents/" .
				$receipt_id;

			// headers
			$headers = ["Authorization: Bearer " . $bifit->access_token];

			// получаем документ
			$doc = json_decode(self::__loadData($url, $headers, false, "GET"));
		}

		// ССЫЛКА НА ЧЕК

		$receipt_link = !is_object($receipt_id)
			? "https://fp" .
				$test .
				".bifit.com/processing-api/receipts/" .
				$idemp_key
			: "XXX";

		// ЛОГ ОПЕРАЦИИ

		// Если включено логирование
		if ($log_ok) {
			// параметры для лога
			$logObj = new stdClass();
			$logObj->order_id = $info["id"];
			$logObj->order_number = $info["cashregister_id"];
			$logObj->order_date = date("Y-m-d H:i:s", $info["created"]);
			$logObj->cashier =
				$cashier_inn != ""
					? $cashier_name . ", ИНН:" . self::__checkInn($cashier_inn)
					: $cashier_name;
			$logObj->client = $address;
			$logObj->receipt_link = $receipt_link;
			$logObj->receipt_id = !is_object($receipt_id) ? $receipt_id : "0";
			$logObj->receipt_type = $receipt_type;
			$logObj->payment_type = $payment_type;
			$logObj->doc = "";
			$logObj->log = "";

			// сохраняем лог операции
			self::__saveReport(
				$test,
				$bifit,
				$receipt,
				$receipt_id,
				$doc,
				$loc,
				$logObj,
				$idemp_key
			);
		}

		return $receipt_id;
	}

	function __formatprc($price)
	{
		$dec_count = $this->diafan->_shop->price_num_decimal_places();

		// форматируем цену согласно настроек
		$price =
			$dec_count < 0
				? round($price, $dec_count)
				: number_format($price, $dec_count, ".", "");

		return $price;
	}

	function __checkInn($inn)
	{
		// только цифры в количестве 12 шт.
		$inn = !preg_match("/^\d{12}$/", $inn)
			? "ИНН задан в неверном формате!"
			: $inn;

		return $inn;
	}

	function __checkPhone($tel)
	{
		// удаляем лишние символы
		$phone = str_replace(["+", "(", ")", "-", " "], "", trim($tel));

		// получаем первую цифру
		$first = substr($phone, "0", 1);

		// если номер из 10 цифр и первая цифра не 7, то добавляем 7
		$phone = strlen($phone) == 10 && $first != 7 ? "7" . $phone : $phone;

		// окончательная проверка
		$phone = !preg_match("/^[0-9]{11}+$/", $phone)
			? "Телефон задан в неверном формате!"
			: $phone;

		return $phone;
	}

	function __checkEmail($email)
	{
		// проверка модели записи
		$email = !preg_match("/^\+\d+$|^\S+@\S+$/", $email)
			? "E-mail задан в неверном формате!"
			: $email;

		return $email;
	}

	function __checkProductPrice(
		$item_price,
		$order_subtotal,
		$order_discount = 0,
		$order_payment = 0
	) {
		// если скидка больше нуля
		if ($order_discount > 0) {
			// вычисляем процент скидки
			$percent = ($order_discount * 100) / $order_subtotal;

			// уменьшаем цену товара согласно скидки
			$item_price = $item_price - $item_price * ($percent / 100);
		}

		// если система оплаты больше нуля
		if ($order_payment > 0) {
			// вычисляем процент наценки
			$percent =
				($order_payment * 100) / ($order_subtotal - $order_discount);

			// увеличиваем цену товара согласно наценки
			$item_price = $item_price + $item_price * ($percent / 100);
		}

		return $item_price;
	}

	function __loadData($url, $headers = false, $data = false, $type = "POST")
	{
		// инициализируем сеанс cURL
		$ch = curl_init($url);

		// назначаем тип запроса
		if ($type != "POST") {
			curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $type);
		} else {
			curl_setopt($ch, CURLOPT_POST, true);
		}

		// добавляем данные
		if ($data) {
			curl_setopt($ch, CURLOPT_POSTFIELDS, $data);
		}

		// возврат результата в качестве строки
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		curl_setopt($ch, CURLOPT_ENCODING, "");
		curl_setopt($ch, CURLOPT_MAXREDIRS, 10);
		curl_setopt($ch, CURLOPT_TIMEOUT, 0);
		curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
		curl_setopt($ch, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);

		// добавляем headers
		if ($headers) {
			curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
		}

		// получаем ответ
		$output = curl_exec($ch);

		// завершаем сеанс cURL
		curl_close($ch);

		return $output;
	}

	function __saveReport(
		$test,
		$bifit,
		$receipt,
		$receipt_id,
		$doc,
		$loc,
		$logObj,
		$idemp_key
	) {
		// выводим место и свойство операции
		$loc = $loc == "admin" ? " (админ)" : " (фронт)";
		$log = "####### Боевой запрос" . $loc . "\n";
		$log = $test != "" ? "####### Тестовый запрос" . $loc . "\n" : $log;

		// устанавливаем часовой пояс согласно настройкам сайта
		$log .= "####### " . date("d-m-Y H:i:s", time()) . "\n";

		// ОТВЕТ #1

		$log .= "####### Ответ #1: " . "\n";

		// данные ответа #1
		$token_error = null;

		if (is_object($bifit)) {
			foreach ($bifit as $key => $val) {
				if ($key == "error") {
					$token_error = 1;
				}
				$log .= $key . ": " . $val . "\n";
			}
		}

		// ЧЕК

		// сформированный чек для отправки
		$log .= "####### Чек: " . "\n" . $receipt . "\n";

		if (!$token_error) {
			// КЛЮЧ ИДЕМПОТЕНТНОСТИ

			$log .= "####### Ключ идемпотентности: " . "\n" . $idemp_key . "\n";

			// ОТВЕТ #2

			$log .= "####### Ответ #2: " . "\n";

			// данные ответа #2
			if (is_object($receipt_id)) {
				$log .= self::__checkResponce($receipt_id);
			} else {
				$log .= "Номер документа: " . $receipt_id . "\n";

				// ОТВЕТ #3

				$log .= "####### Ответ #3: " . "\n";

				if (is_object($doc)) {
					$log .= self::__checkResponce($doc);
				}

				// ССЫЛКА НА ЧЕК

				// сформированная ссылка на чек
				$log .=
					"####### Ссылка на чек: " . "\n" . $logObj->receipt_link;
			}
		}

		$logObj->log = $log;

		// сохранение
		DB::query(
			"INSERT INTO {bifit} (order_id, order_number, order_date,cashier,client,receipt_link,receipt_id,receipt_type,payment_type,doc,log) VALUES (%d,'%s','%s','%s','%s','%s','%s','%s','%s','%h','%h')",
			$logObj->order_id,
			$logObj->order_number,
			$logObj->order_date,
			$logObj->cashier,
			$logObj->client,
			$logObj->receipt_link,
			$logObj->receipt_id,
			$logObj->receipt_type,
			$logObj->payment_type,
			$logObj->doc,
			$logObj->log
		);
	}
}

class BifitException extends Exception
{
}
